#!/usr/bin/env python3

import sys
import argparse
import logging
from typing import Any, Dict
from yabai_stack_navigator.yabai_layout_details import YabaiLayoutDetails
from yabai_stack_navigator.yabai_navigator import YabaiNavigator
from yabai_stack_navigator.yabai_stacked_window_provider \
    import YabaiStackedWindowProvider


def parse_arg_data() -> Dict[str, Any]:
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-n", "--next", action="store_true", help="next window")
    group.add_argument("-p",
                       "--previous",
                       action="store_true",
                       help="previous window")
    parser.add_argument("-v",
                        "--verbose",
                        action="store_true",
                        help="Verbose mode to aid in debugging")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = vars(parser.parse_args())

    if args["verbose"]:
        logging.basicConfig(level=logging.DEBUG)

    return args


def main():
    args = parse_arg_data()

    navigator = YabaiNavigator()

    if YabaiLayoutDetails().is_layout_stacked():
        logging.debug("Stack layout detected")
        window_navigation_data = YabaiStackedWindowProvider(
        ).get_previous_and_next_windows()
        window_key = "next_window" if args['next'] else "previous_window"
        if not args['next'] and not args['previous']:
            raise Exception("Should not get here")
        navigator.focus_on_stacked_window(
            window_navigation_data[window_key]["id"])
    else:
        logging.debug("Non-stacked layout detected")
        if args['next']:
            navigator.focus_on_window()
        elif args['previous']:
            navigator.focus_on_window(next=False)
        else:
            raise Exception("Should not get here")


if __name__ == "__main__":
    main()


import json
import subprocess
import logging
from typing import Any


class YabaiProvider:
    """Class to abstract out calling out to Yabai"""
    def call_yabai(self, args) -> Any:
        if args[0] != "yabai":
            args = ["yabai"] + args

        logging.debug(f"Calling Yabai With Args: {args}")
        command_output = subprocess.run(
            args, stdout=subprocess.PIPE).stdout.decode('utf-8')

        if len(command_output) > 0:
            logging.debug(f"Output from Yabai: {command_output}")
            return json.loads(command_output)
        else:
            logging.debug("No output from Yabai")
            return None



from yabai_stack_navigator.yabai_stacked_window_provider \
     import YabaiStackedWindowProvider
from unittest.mock import Mock


def base_test_data():
    return [
        {
            "focused": 0,
            "id": 1
        },
        {
            "focused": 0,
            "id": 2
        },
        {
            "focused": 0,
            "id": 3
        },
    ]


# Check the scenario when we're navigating from the middle
def test_get_previous_and_next_windows_middle_node():
    mock = Mock()
    test_data = base_test_data()
    test_data[1]["focused"] = 1
    mock.sort_stacked_windows.return_value = test_data
    provider = YabaiStackedWindowProvider(layout_details=mock)

    results = provider.get_previous_and_next_windows()

    assert (results["previous_window"]['id'] == 1)
    assert (results["next_window"]['id'] == 3)


# Check the scenario when we're navigating from the first window
def test_get_previous_and_next_windows_head_node():
    mock = Mock()
    test_data = base_test_data()
    test_data[0]["focused"] = 1
    mock.sort_stacked_windows.return_value = test_data
    provider = YabaiStackedWindowProvider(layout_details=mock)

    results = provider.get_previous_and_next_windows()

    assert (results["previous_window"]['id'] == 3)
    assert (results["next_window"]['id'] == 2)


# Check the scenario when we're navigating from the last window
def test_get_previous_and_next_windows_tail_node():
    mock = Mock()
    test_data = base_test_data()
    test_data[2]["focused"] = 1
    mock.sort_stacked_windows.return_value = test_data
    provider = YabaiStackedWindowProvider(layout_details=mock)

    results = provider.get_previous_and_next_windows()

    assert (results["previous_window"]['id'] == 2)
    assert (results["next_window"]['id'] == 1)


def test_get_previous_and_next_windows_throws_exception():
    mock = Mock()
    test_data = base_test_data()
    mock.sort_stacked_windows.return_value = test_data
    provider = YabaiStackedWindowProvider(layout_details=mock)

    try:
        provider.get_previous_and_next_windows()
    except Exception as e:
        assert (e is not None)
        assert ("Shoudln't" in str(e))



from yabai_stack_navigator.yabai_navigator import YabaiNavigator
from unittest.mock import Mock, patch


def test_focus_on_stacked_window_passes_in_window_id():
    mock = Mock()
    test_window_id = "test_window_id"

    navigator = YabaiNavigator(yabai_provider=mock)
    navigator.focus_on_stacked_window(test_window_id)

    mock.call_yabai.assert_called_with(
        ["-m", "window", "--focus", test_window_id])


@patch('os.system')
def test_focus_on_window_uses_next(mock_os):
    navigator = YabaiNavigator()
    navigator.focus_on_window(next=True)
    mock_os.assert_called_once()
    assert ("next" in mock_os.call_args[0][0])


@patch('os.system')
def test_focus_on_window_uses_previous(mock_os):
    navigator = YabaiNavigator()
    navigator.focus_on_window(next=False)

    mock_os.assert_called_once()
    assert ("prev" in mock_os.call_args[0][0])



from yabai_stack_navigator.yabai_layout_details import YabaiLayoutDetails
from unittest.mock import Mock, MagicMock


def test_get_space_info_uses_correct_parameters():
    mock = Mock()
    layout_details = YabaiLayoutDetails(yabai_provider=mock)

    layout_details.get_space_info()

    mock.call_yabai.assert_called_once_with(
        ["-m", "query", "--spaces", "--space"])


def test_get_data_for_windows_in_space_passes_index():
    mock = Mock()
    layout_details = YabaiLayoutDetails(yabai_provider=mock)
    test_index = "test_index"

    layout_details.get_data_for_windows_in_space(test_index)

    mock.call_yabai.assert_called_once_with(
        ["-m", "query", "--windows", "--space", test_index])


def test_is_layout_stacked():
    data_with_stack = {"type": "stack"}
    data_without_stack = {"type": "window"}

    layout_details = YabaiLayoutDetails()

    layout_details.get_space_info = MagicMock(return_value=data_with_stack)
    assert (layout_details.is_layout_stacked())

    layout_details.get_space_info = MagicMock(return_value=data_without_stack)
    assert (not layout_details.is_layout_stacked())


def test_get_layout_index():
    data = {"index": 5}

    layout_details = YabaiLayoutDetails()

    assert (layout_details.get_layout_index(data) == data["index"])


def test_sort_stack_windows_sorts_by_stack_index():
    test_data = [
        {
            "id": 3,
            "stack-index": 3
        },
        {
            "id": 2,
            "stack-index": 2
        },
        {
            "id": 1,
            "stack-index": 1
        },
    ]

    layout_details = YabaiLayoutDetails()
    results = layout_details.sort_stacked_windows(test_data)

    assert (len(results) == 3)
    assert (results[0]["id"] == 1)
    assert (results[1]["id"] == 2)
    assert (results[2]["id"] == 3)


def test_filter_windows():
    test_data = [
        {
            "window_id": 1,
            "app": "VSCode"
        },
        {
            "window_id": 2,
            "app": "Hammerspoon"
        },
        {
            "window_id": 3,
            "app": "Hammerspoon"
        },
        {
            "window_id": 4,
            "app": "iTerm2"
        },
    ]

    layout_details = YabaiLayoutDetails()
    results = layout_details.filter_windows(test_data)

    assert (len(results) == 2)
    assert (results[0]["window_id"] == 1 and results[0]["app"] == "VSCode")
    assert (results[1]["window_id"] == 4 and results[1]["app"] == "iTerm2")



import logging
from .yabai_provider import YabaiProvider


class YabaiLayoutDetails:
    AppsToFilterOut = set(["Hammerspoon"])

    def __init__(self, yabai_provider=YabaiProvider()):
        self.yabai_provider = yabai_provider

    def get_space_info(self):
        logging.debug("Called get_space_info")
        return self.yabai_provider.call_yabai(
            ["-m", "query", "--spaces", "--space"])

    def get_data_for_windows_in_space(self, index):
        logging.debug("Called get_data_for_windows_in_space")
        return self.yabai_provider.call_yabai(
            ["-m", "query", "--windows", "--space",
             str(index)])

    def is_layout_stacked(self):
        logging.debug("Called is_layout_stacked")
        return self.get_space_info()["type"] == "stack"

    def get_layout_index(self, layout_data):
        logging.debug("Called get_layout_index")
        return layout_data["index"]

    def sort_stacked_windows(self, window_data):
        logging.debug("Called sort_stacked_windows")
        logging.debug(f"Data Before Sort: {window_data}")
        sorted_window_data = sorted(window_data,
                                    key=lambda i: i['stack-index'])
        logging.debug(f"Data After Sort: {sorted_window_data}")
        return sorted_window_data

    def filter_windows(self, window_data):
        logging.debug("Called filter_windows")
        return [
            window for window in window_data
            if window["app"] not in YabaiLayoutDetails.AppsToFilterOut
        ]



import os
import logging
from .yabai_provider import YabaiProvider


class YabaiNavigator:
    def __init__(self, yabai_provider=YabaiProvider()):
        self.yabai_provider = yabai_provider

    def focus_on_stacked_window(self, window_id):
        logging.debug("Focusing on stacked window")
        args = ["-m", "window", "--focus", str(window_id)]
        return self.yabai_provider.call_yabai(args)

    def focus_on_window(self, next=True):
        if next:
            logging.debug("Focusing on next window")
            os.system(
                "yabai -m window --focus stack.next || yabai -m window --focus next || yabai -m window --focus first"
            )
        else:
            logging.debug("Focusing on previous window")
            os.system(
                "yabai -m window --focus stack.prev || yabai -m window --focus prev || yabai -m window --focus last"
            )



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='yabai_stack_navigator',
    packages=find_packages(),
    version='1.0.2',
    description='Script to make navigating between stacks on Yabai easier.',
    author='Sendhil Panchadsaram',
    license='MIT',
    long_description=
    "A simple script to make navigating between stacks and windows in Yabai(https://github.com/koekeishiya/yabai) easier. Details at https://github.com/sendhil/yabai-stack-navigator.",
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    test_suite='tests',
    py_modules=['main', 'yabai_stack_navigator'],
    project_urls={
        'GitHub': 'https://github.com/sendhil/yabai-stack-navigator',
    },
    entry_points='''
        [console_scripts]
        yabai-stack-navigator=main:main
    '''),



from .yabai_layout_details import YabaiLayoutDetails


class YabaiStackedWindowProvider:
    def __init__(self, layout_details=YabaiLayoutDetails()):
        self.layout_details = layout_details

    def get_previous_and_next_windows(self):
        space_data = self.layout_details.get_space_info()
        index = self.layout_details.get_layout_index(space_data)
        all_window_data = self.layout_details.get_data_for_windows_in_space(
            index)
        # This is to primarily remove Hammerspoon which shows up as
        # a part of the layout details in Yabai
        filtered_window_data = self.layout_details.filter_windows(
            all_window_data)
        sorted_window_data = self.layout_details.sort_stacked_windows(
            filtered_window_data)

        number_of_windows = len(sorted_window_data)
        for index, window in enumerate(sorted_window_data):
            if window["focused"] == 1:
                next_window_index = (index + 1) % number_of_windows
                previous_window_index = (index - 1) % number_of_windows
                return {
                    "previous_window":
                    sorted_window_data[previous_window_index],
                    "next_window": sorted_window_data[next_window_index]
                }

        raise Exception("Shoudln't get here")
